﻿#nullable enable

using System;
using System.Collections.Generic;
using Microsoft.Extensions.Configuration;
using Newtonsoft.Json.Linq;
using UnityEngine;

namespace Unity.Extensions.Configuration
{
    public class ResourcesConfigurationProvider : ConfigurationProvider
    {
        private readonly Stack<string> _paths = new();

        public UnityResourcesConfigurationSource Source { get; }

        public ResourcesConfigurationProvider(UnityResourcesConfigurationSource configurationSource)
        {
            Source = configurationSource;
        }

        public override void Load()
        {
            var filePath = Source.Path ?? string.Empty;

            if (string.IsNullOrWhiteSpace(filePath))
            {
                return;
            }

            var textAsset = Resources.Load<TextAsset>(filePath);
            var text = textAsset != null ? textAsset.text : string.Empty;

            if (string.IsNullOrWhiteSpace(text))
            {
                return;
            }

            Parse(text);
        }

        private void Parse(string text)
        {
            var jsonLoadSettings = new JsonLoadSettings()
            {
                CommentHandling = CommentHandling.Ignore
            };

            var jToken = JToken.Parse(text, jsonLoadSettings);

            if (jToken.Root.Type != JTokenType.Object || jToken.Root is not JObject jObject)
            {
                throw new FormatException(
                    $"[{nameof(ResourcesConfigurationProvider)}] Invalid top level JSON element {jToken.Root.Type}");
            }

            ParseObject(jObject);
        }

        private void ParseObject(JObject jObject)
        {
            var isEmpty = true;

            foreach (var property in jObject.Properties())
            {
                isEmpty = false;
                EnterContext(property.Name);
                ParseValue(property.Value);
                ExitContext();
            }

            SetNullIfElementIsEmpty(isEmpty);
        }

        private void ParseValue(JToken value)
        {
            UnityEngine.Assertions.Assert.IsTrue(_paths.Count > 0);
            switch (value.Type)
            {
                case JTokenType.None:
                {
                    // NOP
                    break;
                }
                case JTokenType.Object when value is JObject jObject:
                {
                    ParseObject(jObject);
                    break;
                }
                case JTokenType.Array when value is JArray jArray:
                {
                    ParseArray(jArray);
                    break;
                }
                case JTokenType.Comment:
                {
                    // NOP
                    break;
                }
                case JTokenType.Integer:
                case JTokenType.Float:
                case JTokenType.String:
                case JTokenType.Boolean:
                case JTokenType.Null:
                case JTokenType.Date:
                case JTokenType.Guid:
                case JTokenType.Uri:
                case JTokenType.TimeSpan:
                {
                    var key = _paths.Peek();
                    if (Data.ContainsKey(key))
                    {
                        throw new FormatException($"[{nameof(ResourcesConfigurationProvider)}] Key duplication: {key}");
                    }

                    Data[key] = value.ToString();
                    break;
                }
                case JTokenType.Constructor:
                case JTokenType.Property:
                case JTokenType.Raw:
                case JTokenType.Undefined:
                case JTokenType.Bytes:
                default:
                    throw new FormatException($"[{nameof(ResourcesConfigurationProvider)}] JSON token type {value.Type} isn't supported");
            }
        }

        private void ParseArray(JArray array)
        {
            var index = 0;

            foreach (var jToken in array)
            {
                if (jToken is not JObject jObject)
                {
                    throw new FormatException($"[{nameof(ResourcesConfigurationProvider)}] JSON token type {jToken.Type} isn't supported in an array");
                }
                
                EnterContext(index.ToString());
                ParseValue(jObject);
                ExitContext();
                index++;
            }
            
            SetNullIfElementIsEmpty(index == 0);
        }

        private void SetNullIfElementIsEmpty(bool isEmpty)
        {
            if (isEmpty && _paths.Count > 0)
            {
                Data[_paths.Peek()] = null;
            }
        }

        private void EnterContext(string context)
        {
            _paths.Push(_paths.Count > 0 ? _paths.Peek() + ConfigurationPath.KeyDelimiter + context : context);
        }

        private void ExitContext() => _paths.Pop();
    }
}